package de.starzi.graphics.textures;

import java.io.IOException;
import java.nio.FloatBuffer;

import org.lwjgl.opengl.GL11;
import org.lwjgl.opengl.GL12;

import de.starzi.graphics.opengl.GLError;
import de.starzi.graphics.util.BufferGenerator;
import de.starzi.graphics.util.PNG_Encoder;
import de.starzi.graphics.util.PNG_Encoder.Compression;
import de.starzi.graphics.util.PNG_Encoder.Filter;

/**
 * This class represents a number of pixels in a one-dimensiona, two-dimensional or three-dimensional body, 
 * depending on the texture this image is a part of.
 * 
 * An instance of this class can only exist within a texture. 
 * The only way for the user to get an instance of this class is through its texture.
 * 
 * @author Nic Starzi
 * @date 05.01.2014
 *
 */
public class Image {
	
	private final Texture texture;
	private final int level;
	private int width;
	private int height;
	private int depth;
	
	Image(Texture texture, int level) {
		this.texture = texture;
		this.level = level;
	}
	
	/**
	 * Returns the level at which this image is placed within its texture.
	 */
	public int get_level() {
		return level;
	}
	
	/**
	 * Returns the texture that this image is a part of.
	 */
	public Texture get_texture() {
		return texture;
	}
	
	/**
	 * Width of this image.
	 * When no texel data is loaded width is 0.
	 */
	public int get_width() {
		return width;
	}
	
	/**
	 * Height of this image.
	 * When no texel data is loaded height is 0.
	 * When the texture is a 1D texture height is 1.
	 */
	public int get_heighth() {
		return height;
	}
	
	/**
	 * Depth of this image.
	 * When no texel data is loaded depth is 0.
	 * When the texture is not a 3D texture then depth is 1.
	 */
	public int get_depth() {
		return depth;
	}
	
	/**
	 * Overwrites the texels of this image with the new texels.
	 * Changes width, height and depth of this image accordingly.
	 */
	public void put_texels(Texels texels) {
		if (get_texture().get_type().get_dimension() != texels.get_dimension()) {
			throw new IllegalArgumentException("Texture type and texel dimension in conflict.");
		}
		width = texels.get_width();
		height = texels.get_height();
		depth = texels.get_depth();
		
		get_texture().get_type().image(get_level(), width, height, depth, 
				texels.get_buffer(), texels.get_dataType(), texels.get_originalFormat(), texels.get_targetFormat());
	}
	
	/**
	 * Overwrites a part of the texels of this image.
	 * Width, height and depth of this image is not changed by this method.
	 */
	public void put_texels(Texels texels, int offsetX, int offsetY, int offsetZ) {
		if (get_texture().get_type().get_dimension() != texels.get_dimension()) {
			throw new IllegalArgumentException("Texture type and texel dimension in conflict.");
		}
		
		get_texture().get_type().subImage(get_level(), texels.get_width(), texels.get_height(), texels.get_depth(), 
				offsetX, offsetY, offsetZ, texels.get_buffer(), texels.get_dataType(), 
				texels.get_originalFormat(), texels.get_targetFormat());
	}
	
	/**
	 * Saves this image as a .png file using the PNG_Encoder.
	 * Uses the given filter and compression.
	 */
	public void saveToDisk(String path, Filter filter, Compression compression) throws IOException {
		PNG_Encoder.save(path, get_pixels(), width, height, filter, compression);
	}
	
	/**
	 * Equal to:
	 * 	get_pixels ( TexelFormat.RGBA )
	 */
	public float[] get_pixels() {
		return get_pixels(TexelFormat.RGBA);
	}
	
	/**
	 * Returns an array of all texels in this image.
	 * This is a costly operation since the texels are not stored on client side.
	 * The texture must be bound before pixels can be retrieved.
	 */
	public float[] get_pixels(TexelFormat format) {
		if (!get_texture().is_bound()) {
			throw new IllegalStateException("Texture must be bound.");
		}
		int bufferSize = get_width() * get_heighth() * get_depth();
		
		float[] pixels = new float[bufferSize * format.get_byteSize()];
		
		FloatBuffer buf = BufferGenerator.make(pixels);
//		GL11.glGetTexImage(GL11.GL_TEXTURE_2D, get_level(), format.get_glEnum(), GL11.GL_FLOAT, buf);
		GL11.glGetTexImage(get_texture().get_type().get_glEnum(), get_level(), format.get_glEnum(), GL11.GL_FLOAT, buf);
		GLError.throwIfNecessary(this, "glGetTexImage");
		buf.get(pixels);
		
		return pixels;
	}
	
	/**
	 * Helper method used internally when mipmaps are created by the gl.
	 */
	void generateFromMipmap() {
		width = get_texLevelParameter_i(GL11.GL_TEXTURE_WIDTH);
		height = get_texLevelParameter_i(GL11.GL_TEXTURE_HEIGHT);
		depth = get_texLevelParameter_i(GL12.GL_TEXTURE_DEPTH);
	}
	
	public float get_texLevelParameter_f(int pname) {
		float result = GL11.glGetTexLevelParameterf(get_texture().get_type().get_glEnum(), get_level(), pname);
		GLError.throwIfNecessary(this, "glGetTexLevelParameterf");
		return result;
	}
	
	public int get_texLevelParameter_i(int pname) {
		int result = GL11.glGetTexLevelParameteri(get_texture().get_type().get_glEnum(), get_level(), pname);
		GLError.throwIfNecessary(this, "glGetTexLevelParameteri");
		return result;
	}
	
	public String toString() {
		return "["+get_level()+"] "+get_width()+"x"+get_heighth()+"x"+get_depth();
	}
	
}