package de.starzi.graphics.textures;

import org.lwjgl.opengl.GL11;

import de.starzi.graphics.opengl.GLEnum;

/**
 * The texture minifying function is used whenever the level-of-detail function used when 
 * sampling from the texture determines that the texture should be minified.
 * 
 * @author Nic Starzi
 * @date 29.12.2013
 *
 */
public enum MinFilter implements GLEnum {
	
	/**
	 * Returns the value of the texture element that is nearest (in Manhattan distance) to the specified texture coordinates.
	 * GL_NEAREST is generally faster than GL_LINEAR, but it can produce textured images with sharper edges because the transition between texture elements is not as smooth.
	 */
	NEAREST					(GL11.GL_NEAREST),
	
	/**
	 * Chooses the mipmap that most closely matches the size of the pixel being textured and uses the GL_NEAREST criterion 
	 * (the texture element closest to the specified texture coordinates) to produce a texture value.
	 */
	NEAREST_MIPMAP_NEAREST	(GL11.GL_NEAREST_MIPMAP_NEAREST),
	
	/**
	 * Chooses the mipmap that most closely matches the size of the pixel being textured and uses the GL_LINEAR criterion 
	 * (a weighted average of the four texture elements that are closest to the specified texture coordinates) to produce a texture value.
	 */
	NEAREST_MIPMAP_LINEAR	(GL11.GL_NEAREST_MIPMAP_LINEAR),
	
	/**
	 * Returns the weighted average of the texture elements that are closest to the specified texture coordinates. 
	 * These can include items wrapped or repeated from other parts of a texture, depending on the values of GL_TEXTURE_WRAP_S and GL_TEXTURE_WRAP_T, and on the exact mapping.
	 */
	LINEAR					(GL11.GL_LINEAR),
	
	/**
	 * Chooses the two mipmaps that most closely match the size of the pixel being textured and uses the GL_NEAREST criterion 
	 * (the texture element closest to the specified texture coordinates ) to produce a texture value from each mipmap. 
	 * The final texture value is a weighted average of those two values.
	 */
	LINEAR_MIPMAP_NEAREST	(GL11.GL_LINEAR_MIPMAP_NEAREST),
	
	/**
	 * Chooses the two mipmaps that most closely match the size of the pixel being textured and uses the GL_LINEAR criterion 
	 * (a weighted average of the texture elements that are closest to the specified texture coordinates) to produce a texture value from each mipmap. 
	 * The final texture value is a weighted average of those two values.
	 */
	LINEAR_MIPMAP_LINEAR	(GL11.GL_LINEAR_MIPMAP_LINEAR);
	
	private final int glEnum;
	
	private MinFilter(int glEnum) {
		this.glEnum = glEnum;
	}
	
	public int get_glEnum() {
		return glEnum;
	}
	
}