package de.starzi.graphics.textures;

import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.nio.ByteBuffer;

import de.matthiasmann.twl.utils.PNGDecoder;
import de.matthiasmann.twl.utils.PNGDecoder.Format;
import de.starzi.graphics.opengl.GLDataType;
import de.starzi.graphics.util.BufferGenerator;

/**
 * This class represents a number of Pixels as they are loaded from an image on disk, or texels from a texture.
 * To create an instance of this class use the static methods it provides.
 * 
 * This class is most commonly used to upload pixels into a texture.
 * 
 * @author Nic Starzi
 * @date 05.01.2014
 *
 */
public class Texels {
	
	public static final TexelInternalFormat DEFAULT_INTERNAL_FORMAT = TexelInternalFormat.RGBA8;
	
	/**
	 * Constructs texels from the given float array using the given TexelFormat.
	 * The following condition must be true:
	 * 		pixels.length() == width * height * depth
	 * 
	 * If depth and height are 1 the texels will be considered one-dimensional.
	 * If depth is 1 and height is greater then 1 the texels will be considered two-dimensional.
	 * If depth is greater then 1 the texels will be considered three-dimensional.
	 */
	public static Texels image(int width, int height, int depth, float[] pixels, TexelFormat incomingFormat) {
		int dimension = 1;
		if (height > 1) {
			dimension = 2;
		}
		if (depth > 1) {
			dimension = 3;
		}
		
		return new Texels(width, height, depth, dimension, 
				BufferGenerator.make_byteBuffer(pixels), 
				GLDataType.FLOAT, incomingFormat);
	}
	
	/**
	 * Loads a .png image from disk and generates 1D texels for it.
	 */
	public static Texels image1D(String path) throws IOException {
		InputStream stream = new FileInputStream(path);
		return image1D(stream);
	}
	
	/**
	 * Loads a .png image from the stream and generates 1D texels for it.
	 */
	public static Texels image1D(InputStream stream) throws IOException {
		PNGDecoder decoder = new PNGDecoder(stream);
		
		if (decoder.getHeight() != 1) {
			stream.close();
			throw new RuntimeException("1D image needs to have a height of exactly 1 pixel. Was: "+decoder.getHeight());
		}
		
		int w = get_power_of_two(decoder.getWidth());
		int h = 1;
		int d = 1;
		GLDataType dataType = GLDataType.BYTE;
		TexelFormat incomingFormat = TexelFormat.BGRA;
		
		ByteBuffer buffer = BufferGenerator.make_byteBuffer(w * h * d * incomingFormat.get_byteSize()); 
		decoder.decode(buffer, decoder.getWidth() * 4, get_format(incomingFormat));
		while (buffer.remaining() > 0){
			buffer.put(Byte.MIN_VALUE);
		}
		buffer.flip();
		stream.close();
		
		return new Texels(w, h, d, Type._1D, buffer, dataType, incomingFormat);
	}
	
	/**
	 * Loads a .png image from disk and generates 2D texels for it.
	 */
	public static Texels image2D(String path) throws IOException {
		InputStream stream = new FileInputStream(path);
		return image2D(stream);
	}
	
	/**
	 * Loads a .png image from the stream and generates 2D texels for it.
	 */
	public static Texels image2D(InputStream stream) throws IOException {
		PNGDecoder decoder = new PNGDecoder(stream);
		
		int w = get_power_of_two(decoder.getWidth());
		int h = get_power_of_two(decoder.getHeight());
		int d = 1;
		GLDataType dataType = GLDataType.BYTE;
		TexelFormat incomingFormat = TexelFormat.RGBA;
		
		ByteBuffer buffer = BufferGenerator.make_byteBuffer(w * h * d * incomingFormat.get_byteSize()); 
		decoder.decode(buffer, decoder.getWidth() * 4, get_format(incomingFormat));
		while (buffer.remaining() > 0){
			buffer.put(Byte.MIN_VALUE);
		}
		buffer.flip();
		stream.close();
		
		return new Texels(w, h, d, Type._2D, buffer, dataType, incomingFormat);
	}
	
	/**
	 * Loads a .png image from disk and generates 3D texels for it.
	 */
	public static Texels image3D(String[] paths) throws IOException {
		InputStream[] streams = new InputStream[paths.length];
		for (int i = 0; i < paths.length; i++) {
			streams[i] = new FileInputStream(paths[i]);
		}
		return image3D(streams);
	}
	
	/**
	 * Loads a .png image from the stream and generates 3D texels for it.
	 */
	public static Texels image3D(InputStream[] streams) throws IOException {
		int width = 0;
		int height = 0;
		
		Texels[] layers = new Texels[streams.length];
		for (int i = 0; i < layers.length; i++) {
			layers[i] = image2D(streams[i]);
			if (i == 0) {
				width = layers[i].width;
				height = layers[i].height;
			} else if (width != layers[i].width
					|| height  != layers[i].height) {
				throw new RuntimeException("Each image must have the same size.");
			}
		}
		
		int w = width;
		int h = height;
		int d = layers.length;
		GLDataType dataType = GLDataType.BYTE;
		TexelFormat incomingFormat = TexelFormat.BGRA;
		
		ByteBuffer buffer = BufferGenerator.make_byteBuffer(w * h * d * incomingFormat.get_byteSize());
		for (int i = 0; i < layers.length; i++) {
			buffer.put(layers[i].buffer);
		}
		while (buffer.remaining() > 0){
			buffer.put(Byte.MIN_VALUE);
		}
		buffer.flip();
		
		return new Texels(w, h, d, Type._3D, buffer, dataType, incomingFormat);
	}
	
	private final int width;
	private final int height;
	private final int depth;
	private final ByteBuffer buffer;
	private final int dimension;
	private final GLDataType dataType;
	private final TexelFormat incomingFormat;
	private TexelInternalFormat internalFormat;
	
	private Texels(int w, int h, int d, Type type, ByteBuffer buffer, GLDataType dataType, TexelFormat incoming) {
		this(w, h, d, type.get_dimension(), buffer, dataType, incoming);
	}
	
	private Texels(int w, int h, int d, int dim, ByteBuffer buffer, GLDataType dataType, TexelFormat incoming) {
		this.width = w;
		this.height = h;
		this.depth = d;
		this.dimension = dim;
		this.buffer = buffer;
		this.dataType = dataType;
		this.incomingFormat = incoming;
		this.internalFormat = DEFAULT_INTERNAL_FORMAT;
	}
	
	public int get_dimension() {
		return dimension;
	}
	
	public int get_width() {
		return width;
	}
	
	public int get_height() {
		return height;
	}
	
	public int get_depth() {
		return depth;
	}
	
	public ByteBuffer get_buffer() {
		return buffer;
	}
	
	public GLDataType get_dataType() {
		return dataType;
	}
	
	public TexelFormat get_originalFormat() {
		return incomingFormat;
	}
	
	public void set_targetFormat(TexelInternalFormat format) {
		internalFormat = format;
	}
	
	public TexelInternalFormat get_targetFormat() {
		return internalFormat;
	}
	
	private static int get_power_of_two(int value) {
		return is_power_of_two(value) ? value : next_power_of_two(value);
	}
	
	private static int next_power_of_two(int value){
		value--;
		value |= value >> 1;
		value |= value >> 2;
		value |= value >> 4;
		value |= value >> 8;
		value |= value >> 16;
		value++;
		return value;
	}
	
	private static boolean is_power_of_two(final int value){
		return (value & (~value + 1)) == value;
	}
	
	@SuppressWarnings("incomplete-switch")
	private static PNGDecoder.Format get_format(TexelFormat format) {
		switch (format) {
		case BGRA:
			return Format.BGRA;
		case RED:
			return Format.ALPHA;
		case RGB:
			return Format.RGB;
		case RGBA:
			return Format.RGBA;
		}
		throw new RuntimeException("Format is not supported by PNGDecoder: "+format);
	}
	
}