package de.starzi.graphics.textures;

import java.util.Arrays;

import org.lwjgl.opengl.GL11;
import org.lwjgl.opengl.GL12;
import org.lwjgl.opengl.GL14;
import org.lwjgl.opengl.GL30;

import de.starzi.Bindable;
import de.starzi.Disposable;
import de.starzi.graphics.opengl.GLError;
import de.starzi.graphics.opengl.GLHint;
import de.starzi.graphics.opengl.GLObject;

/**
 * This class is a representation of an OpenGL texture.
 * A texture consists of a number of images and a state.
 * 
 * A texture always has a constant texture-type which does never change after the texture has been created.
 * 
 * @author Nic Starzi
 * @date 05.01.2014
 *
 */
public class Texture implements GLObject, Bindable, Disposable {
	
	public static void disable_textures() {
		Type.disable_all();
	}
	
	private final Type type;
	private Image[] images;
	private ImageUnit textureImageUnit;
	
	private final int glName;
	private int baseLevel;
	private boolean disposed;
	
	/**
	 * Creates a new texture and sets its type.
	 * The type of the texture does never change after it has been created.
	 */
	public Texture(Type type) {
		this.glName = GL11.glGenTextures();
		GLError.throwIfNecessary(this, "glGenTextures");
		this.type = type;
		this.textureImageUnit = ImageUnit.get(0);
	}
	
	public void set_minFilter(MinFilter filter) {
		set_texParameter_i(GL11.GL_TEXTURE_MIN_FILTER, filter.get_glEnum());
	}
	
	public void set_magFilter(MagFilter filter) {
		set_texParameter_i(GL11.GL_TEXTURE_MAG_FILTER, filter.get_glEnum());
	}
	
	/**
	 * Horizontal wrap.
	 */
	public void set_wrap_s(Wrap wrap) {
		set_texParameter_i(GL11.GL_TEXTURE_WRAP_S, wrap.get_glEnum());
	}
	
	/**
	 * Vertical wrap. Only needed for 2D, and 3D textures.
	 */
	public void set_wrap_t(Wrap wrap) {
		set_texParameter_i(GL11.GL_TEXTURE_WRAP_T, wrap.get_glEnum());
	}
	
	/**
	 * Wrap across third axis. Only needed for 3D textures.
	 */
	public void set_wrap_r(Wrap wrap) {
		set_texParameter_i(GL12.GL_TEXTURE_WRAP_R, wrap.get_glEnum());
	}
	
	/**
	 * Specifies the lowest and highest mipmap levels for this texture.
	 * Mipmaps will not be generated though, you need to either call 
	 * generate_mipmaps() or specify the levels yourself via texel upload.
	 * 
	 * baseLevel must be >= 0
	 * maxLevel must be >= baseLevel
	 */
	public void set_levels(int baseLevel, int maxLevel) {
		if (baseLevel < 0) {
			throw new IllegalArgumentException("Base level must not be negative.");
		}
		if (maxLevel < baseLevel) {
			throw new IllegalArgumentException("Max level must not be less then base level.");
		}
		this.baseLevel = baseLevel;
		
		set_texParameter_i(GL12.GL_TEXTURE_BASE_LEVEL, baseLevel);
		set_texParameter_i(GL12.GL_TEXTURE_MAX_LEVEL, maxLevel);
		
		int newSize = (maxLevel - baseLevel) + 1;
		if (images == null) {
			images = new Image[newSize];
		} else {
			images = Arrays.copyOfRange(images, baseLevel, maxLevel);
		}
		for (int i = 0; i < images.length; i++) {
			if (images[i] == null) {
				images[i] = new Image(this, i);
			}
		}
	}
	
	/**
	 * Returns the image at the given level.
	 * Following condition must be true:
	 * 	level >= get_baseLevel() && level < get_maxLevel()
	 * Otherwise will throw ArrayOutOfBoundsException
	 */
	public Image get_image(int level) {
		return images[level - baseLevel];
	}
	
	/**
	 * The minimum mipmap level for this texture.
	 */
	public int get_baseLevel() {
		return baseLevel;
	}
	
	/**
	 * The highest mipmap level for this texture. 
	 */
	public int get_maxLevel() {
		return images.length + get_baseLevel() - 1;
	}
	
	/**
	 * Returns the maximum number of mipmaps this texture can generate automatically.
	 * Any more mipmaps then this number would not improve the quality of mipmapping.
	 * 
	 * The parameters should be the dimension of the base level for the texture.
	 * 
	 * Please note that the number of mipmaps is not equal to the max level of the texture.
	 * The max level is: baseLevel + mipmapCount - 1
	 */
	public int get_recommendedLevelCount(int textureWidth, int textureHeight, int textureDepth) {
		int biggestDimension = Math.max(Math.max(textureWidth, textureHeight), textureDepth);
		return 1 + (int) Math.ceil(Math.log(biggestDimension) / Math.log(2));
	}
	
	/**
	 * Generates the mipmaps for this texture automatically.
	 * You have to set the min level and max level apropriately before calling this.
	 */
	public void generate_mipmaps(GLHint hint) {
		hint.perform(GL14.GL_GENERATE_MIPMAP_HINT);
		
		GL30.glGenerateMipmap(get_type().get_glEnum());
		GLError.throwIfNecessary(this, "glGenerateMipmap");
		
		for (int i = 1; i < images.length; i++) {
			images[i].generateFromMipmap();
		}
	}
	
	/**
	 * The type of a texture does never change after its initialization.
	 */
	public Type get_type() {
		return type;
	}
	
	/**
	 * The name of a texture does never change after its initialization.
	 */
	public int get_glName() {
		return glName;
	}
	
	public void dispose() {
		throwExceptionWhenDisposed();
		
		if (is_bound()) {
			unbind();
		}
		
		disposed = true;
		GL11.glDeleteTextures(get_glName());
		GLError.throwIfNecessary(this, "glDeleteTextures");
	}
	
	public boolean is_disposed() {
		return disposed;
	}
	
	/**
	 * Binds the texture to a specific image unit rather then the default one / last one used.
	 */
	public void bind(ImageUnit textureImageUnit) {
		throwExceptionWhenDisposed();
		
		if (!get_type().is_enabled()) {
			get_type().enable();
		}
		ImageUnit currentTextureUnit = get_textureImageUnit();
		if (is_bound() && currentTextureUnit.equals(textureImageUnit) ) {
			return;
		}
		
		if (is_bound()) {
			currentTextureUnit.activate();
			unbind();
		}
		textureImageUnit.activate();
		get_type().set_bound(this);
		this.textureImageUnit = textureImageUnit;
	}
	
	public void bind() {
		bind(textureImageUnit);
	}
	
	/**
	 * Each texture can only be bound to one image unit at a time.
	 * This call will never return null.
	 * When the texture is not bound at all it will return the last image unit used.
	 * When this texture was never ever bound it will return the default image unit.
	 */
	public ImageUnit get_textureImageUnit() {
		return textureImageUnit;
	}
	
	public void unbind() {
		throwExceptionWhenNotReady();
		get_type().set_bound(null);
	}
	
	public boolean is_bound() {
		return equals(get_type().get_bound());
	}
	
	/**
	 * Wrapper around openGL function glTexParameteri.
	 * This method should be avoided in favor of the more precise texture methods.
	 * This method does not check for correctness of either pName or value.
	 * 
	 * @param pName - the constant integer value defining the glEnum.
	 * @param value
	 */
	public void set_texParameter_i(int pName, int value) {
		throwExceptionWhenNotReady();
		GL11.glTexParameteri(get_type().get_glEnum(), pName, value);
		GLError.throwIfNecessary(this, "glTexParameteri");
	}
	
	/**
	 * Wrapper around openGL function glGetTexParameterf.
	 * This method should be avoided in favor of the more precise texture methods if available.
	 */
	public float get_texParameter_f(int pName) {
		float result = GL11.glGetTexParameterf(get_type().get_glEnum(), pName);
		GLError.throwIfNecessary(this, "glGetTexParameterf");
		return result;
	}
	
	/**
	 * Wrapper around openGL function glTexParameterf.
	 * This method should be avoided in favor of the more precise texture methods.
	 * This method does not check for correctness of either pName or value.
	 * 
	 * @param pName - the constant integer value defining the glEnum.
	 * @param value
	 */
	public void set_texParameter_f(int pName, float value) {
		throwExceptionWhenNotReady();
		GL11.glTexParameterf(get_type().get_glEnum(), pName, value);
		GLError.throwIfNecessary(this, "glTexParameterf");
	}
	
	/**
	 * Wrapper around openGL function glGetTexParameteri.
	 * This method should be avoided in favor of the more precise texture methods if available.
	 */
	public int get_texParameter_i(int pName) {
		int result = GL11.glGetTexParameteri(get_type().get_glEnum(), pName);
		GLError.throwIfNecessary(this, "glGetTexParameteri");
		return result;
	}
	
	/**
	 * Combines throwExceptionWhenNotBound and throwExceptionWhenDisposed in one handy dandy call.
	 */
	private void throwExceptionWhenNotReady() {
		throwExceptionWhenNotBound();
		throwExceptionWhenDisposed();
	}
	
	/**
	 * Does exactly what it says on the tin.
	 * Exception thrown is an IllegalStateException
	 */
	private void throwExceptionWhenNotBound() {
		if (!is_bound()) {
			throw new IllegalStateException("Texture is not bound.");
		}
	}
	
	/**
	 * Does exactly what it says on the tin.
	 * Exception thrown is an IllegalStateException
	 */
	private void throwExceptionWhenDisposed() {
		if (is_disposed()) {
			throw new IllegalStateException("Disposed");
		}
	}
	
}