package de.starzi.graphics.textures;

import java.nio.ByteBuffer;

import org.lwjgl.opengl.GL11;
import org.lwjgl.opengl.GL12;
import org.lwjgl.opengl.GL30;

import de.starzi.Activatable;
import de.starzi.graphics.opengl.GLDataType;
import de.starzi.graphics.opengl.GLEnum;
import de.starzi.graphics.opengl.GLError;

/**
 * 
 * @author Nic Starzi
 * @date 05.01.2014
 *
 */
public enum Type implements GLEnum, Activatable {
	
	_1D			(GL11.GL_TEXTURE_1D, 1) {
		public void subImage(int level, int w, int h, int d, int offsetX,
				int offsetY, int offsetZ, ByteBuffer buffer,
				GLDataType dataType, TexelFormat incoming,
				TexelInternalFormat target) {
			
			GL11.glTexSubImage1D(get_glEnum(), level, offsetX, w, incoming.get_glEnum(), dataType.get_glEnum(), buffer);
			GLError.throwIfNecessary(this, "glTexSubImage1D");
		}
		public void image(int level, int w, int h, int d, ByteBuffer buffer,
				GLDataType dataType, TexelFormat incoming,
				TexelInternalFormat target) {
			
			int border = 0;
			GL11.glTexImage1D(get_glEnum(), level, target.get_glEnum(), w, border, incoming.get_glEnum(), dataType.get_glEnum(), buffer);
			GLError.throwIfNecessary(this, "glTexImage1D");
		}
	},
	_1D_ARRAY	(GL30.GL_TEXTURE_1D_ARRAY, 2) {
		public void subImage(int level, int w, int h, int d, int offsetX,
				int offsetY, int offsetZ, ByteBuffer buffer,
				GLDataType dataType, TexelFormat incoming,
				TexelInternalFormat target) {
		}
		public void image(int level, int w, int h, int d, ByteBuffer buffer,
				GLDataType dataType, TexelFormat incoming,
				TexelInternalFormat target) {
		}
	},
	_2D			(GL11.GL_TEXTURE_2D, 2) {
		public void subImage(int level, int w, int h, int d, int offsetX,
				int offsetY, int offsetZ, ByteBuffer buffer,
				GLDataType dataType, TexelFormat incoming,
				TexelInternalFormat target) {
			
			GL11.glTexSubImage2D(get_glEnum(), level, offsetX, offsetY, w, h, incoming.get_glEnum(), dataType.get_glEnum(), buffer);
			GLError.throwIfNecessary(this, "glTexSubImage2D");
		}
		public void image(int level, int w, int h, int d, ByteBuffer buffer,
				GLDataType dataType, TexelFormat incoming,
				TexelInternalFormat target) {
			
			int border = 0;
			GL11.glTexImage2D(get_glEnum(), level, target.get_glEnum(), w, h, border, incoming.get_glEnum(), dataType.get_glEnum(), buffer);
			GLError.throwIfNecessary(this, "glTexImage2D");
		}
	},
	_2D_ARRAY	(GL30.GL_TEXTURE_2D_ARRAY, 3) {
		public void subImage(int level, int w, int h, int d, int offsetX,
				int offsetY, int offsetZ, ByteBuffer buffer,
				GLDataType dataType, TexelFormat incoming,
				TexelInternalFormat target) {
		}
		public void image(int level, int w, int h, int d, ByteBuffer buffer,
				GLDataType dataType, TexelFormat incoming,
				TexelInternalFormat target) {
		}
	},
	_3D			(GL12.GL_TEXTURE_3D, 3) {
		public void subImage(int level, int w, int h, int d, int offsetX,
				int offsetY, int offsetZ, ByteBuffer buffer,
				GLDataType dataType, TexelFormat incoming,
				TexelInternalFormat target) {
			
			GL12.glTexSubImage3D(get_glEnum(), level, offsetX, offsetY, offsetZ, w, h, d, incoming.get_glEnum(), dataType.get_glEnum(), buffer);
			GLError.throwIfNecessary(this, "glTexSubImage3D");
		}
		public void image(int level, int w, int h, int d, ByteBuffer buffer,
				GLDataType dataType, TexelFormat incoming,
				TexelInternalFormat target) {
			
			int border = 0;
			GL12.glTexImage3D(get_glEnum(), level, target.get_glEnum(), w, h, d, border, incoming.get_glEnum(), dataType.get_glEnum(), buffer);
			GLError.throwIfNecessary(this, "glTexImage3D");
		}
	};
	
	public static void disable_all() {
		for (int i = 0; i < values().length; i++) {
			values()[i].disable();
		}
	}
	
	private final int glEnum;
	private final int dimension;
	private boolean enabled;
	private Texture bound;
	
	private Type(int glEnum, int dimension) {
		this.glEnum = glEnum;
		this.dimension = dimension;
	}
	
	public abstract void subImage(int level, int w, int h, int d,
			int offsetX, int offsetY, int offsetZ,
			ByteBuffer buffer, GLDataType dataType, TexelFormat incoming, TexelInternalFormat target);
	
	public abstract void image(int level, int w, int h, int d,
			ByteBuffer buffer, GLDataType dataType, TexelFormat incoming, TexelInternalFormat target);
	
	public int get_glEnum() {
		return glEnum;
	}
	
	public int get_dimension() {
		return dimension;
	}
	
	public void enable() {
		if (!is_enabled()) {
			GL11.glEnable(get_glEnum());
			GLError.throwIfNecessary(this, "glEnable");
			
			enabled = true;
		}
	}
	
	public void disable() {
		if (is_enabled()) {
			GL11.glDisable(get_glEnum());
			GLError.throwIfNecessary(this, "glDisable");
			
			enabled = false;
		}
	}
	
	public boolean is_enabled() {
		return enabled;
	}
	
	public void set_bound(Texture texture) {
		if (texture == null) {
			bound = null;
			bind(0);
			
		} else if (texture.get_type().equals(this)) {
			this.bound = texture;
			bind(texture.get_glName());
			
		} else {
			throw new IllegalArgumentException("Texture dimension is incompatible.");
		}
	}
	
	public Texture get_bound() {
		return bound;
	}
	
	private void bind(int glName) {
		GL11.glBindTexture(get_glEnum(), glName);
		GLError.throwIfNecessary(this, "glBindTexture");
	}
	
}